/*--

   Copyright (c) 2004 Ambros Marzetta and David Bremner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

--*/
/* nq_main.c

   n queens: main program

   Ambros Marzetta

   n	Number of solutions
   3	0
   4	2
   5	10
   6	4
   7	40
   8	92
   9	352
   10	724
   11	2680
   12	14200
   13	73712
   14	365596

 * $Source$
 * $Revision: 5167 $
 * $Date: 2004-10-20 13:20:33 -0300 (Wed, 20 Oct 2004) $
 * $Log$
 * Revision 1.3  2004/10/20 16:20:31  bremner
 * added copyright notice and GPL boilerplate
 *
 * Revision 1.2  2000/11/15 16:14:51  bremner
 *
 * committing patched version of zram source
 *
 * Revision 1.3  1998/04/15  18:34:21  marzetta
 * The program now counts the solutions (#define COUNT).
 *
 * Revision 1.2  1998/03/27  16:47:19  marzetta
 * Fixed a syntax error with an empty statement.
 *
 * Revision 1.1  1998/03/10  15:03:35  marzetta
 * Initial revision
 *
 */

/* Define the following variable if you want the program to
   output the number of solutions found to stderr. */
#define COUNT

#include <stdlib.h>		/* exit() */
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>
#include "z_base.h"
#include "z_bt.h"
#include "z_enumps.h"
#include "z_par.h"		/* z_myproc */
#include "z_time.h"
#ifdef COUNT
#include "z_lb.h"		/* z_lb_redundantComputation */
#endif /* COUNT */

#define MaxN 35

enum outputType {ot_normal, characterGraphic, postScript, noOutput};

static int nq_NodeTag, nq_GlobalTag;

typedef struct nq_Node { /* opaque, a node of the tree to be searched */
  Z_NODE_HEADER;                /* len; tag = nq_NodeTag */
  int d;			/* 0 <= d <= n */
  int p[MaxN];
  int row_occupied[MaxN], diag1_occupied[2*MaxN], diag2_occupied[2*MaxN];
  /* p[0..d-1] are in use */
  /* p[0..d-2] are consistent */
} nq_Node, *nq_NodeP;

typedef struct nq_Global {
  Z_NODE_HEADER;                /* len; tag = nq_GlobalTag */
  int n;			/* must be > 0 */
  enum outputType outputType;
} nq_Global, *nq_GlobalP;

/* Static variables --------------------------------------------------- */

static z_bt_ProcSetT nq_procset;
static int nq_procsetnum;

/* If counting the solutions is desired ------------------------------- */

#ifdef COUNT
static int nq_TotalTag;

typedef struct nq_Total {
  Z_NODE_HEADER;
  long sol_count;		/* count of solutions */
} nq_Total, *nq_TotalP;

static nq_TotalP total;

#ifndef SEQUENTIAL
static z_JobNum totalJobNum;

void CollectTotalJob(void)
{
  /* This procedure, which is called on every processor at the end of the
     computation, sends the local count of solutions back to the front end.
     */
  z_AddNodeBuffer((z_NodeP)total);
} /* CollectTotalJob */

void AddTotal(int pe)
{
  /* This procedure is executed on the front end. It is called for
     every message received from a CollectTotalJob. */
  nq_TotalP t;
  
  t = (nq_TotalP) z_GetNodeBuffer();
  if (pe != z_myproc) {
    total->sol_count += t->sol_count;
  }
  z_FreeNode((z_NodeP) t);
} /* AddTotal */
#endif /* SEQUENTIAL */

#endif /* COUNT */

/* Procedures --------------------------------------------------------- */

static nq_NodeP nq_NewNode(void)
{
  return (nq_NodeP) z_NewNode(nq_NodeTag, sizeof(nq_Node));
} /* nq_NewNode */

#if 0
static void nq_FreeNode(nq_NodeP node)
{
  z_FreeNode((z_NodeP) node);
} /* nq_FreeNode */
#endif

static void nq_Print(nq_GlobalP global, nq_NodeP node)
{
  int i;
  
  for (i=0;i<node->d;i++) printf("%3ld", (long)node->p[i]);
  printf("\n");
} /* nq_Print */

void nq_PSHeader(nq_GlobalP global, FILE *psfile, float xsize, float ysize)
     /* PostScript output: called by z_PrintPS */
{

  fprintf(psfile, "/size %f def\n", xsize); /* size of picture */
  fprintf(psfile, "/small %f def\n", xsize / global->n);
				/* size of little square */
  fprintf(psfile, "/grid {\n"
	  "  1 1 %d {newpath small mul 0 moveto 0 size rlineto stroke} for\n"
	  "  1 1 %d {newpath small mul 0 exch moveto"
	  " size 0 rlineto stroke} for\n"
	  "} def\n",
	  global->n-1, global->n-1);
  fprintf(psfile, "/radius %f def\n",
	  xsize / global->n * 0.3);
  fprintf(psfile, "/queen {0.5 add %f mul exch dup %f mul 3 2 roll "
	  "newpath radius 0 360 arc fill 1 add} def\n",
	  xsize / global->n, ysize / global->n);
} /* nq_PSHeader */

void nq_PSPicture(nq_GlobalP global, nq_NodeP node, FILE *psfile,
		  float xsize, float ysize)
     /* PostScript output: called by z_PrintPS */
{
  int i;
  
  fprintf(psfile, "grid 0.5 [\n");	/* defined in nq_PSHeader */
  for (i=0; i<node->d; i++)
    fprintf(psfile, "%d ", node->p[i]);
  fprintf(psfile, "] {queen} forall pop\n");
} /* nq_PSPicture */

static void nq_Solution(nq_GlobalP global, nq_NodeP node)
{
#ifdef COUNT
  /* count the solutions */
  if (!z_lb_redundantComputation) total->sol_count ++;
#endif /* COUNT */
  
  switch (global->outputType) {
  default:
    nq_Print(global, node);
    break;
  case postScript:
    z_Save((z_NodeP) global, (z_NodeP) node);
    break;
  case noOutput:
    break;
  }
} /* nq_Solution */

boolean nq_BranchInPlace(nq_GlobalP global, nq_NodeP node,
			 z_bt_DoProc whatToDo, void *ref)
{
  int col, row, *p;

  col = node->d;
  /*printf("nq_BranchInPlace %d %d\n", col, global->n);*/
  p = &node->p[col];
  node->d++;
  for (row = global->n-1; row>=0; --row)
    if (! node->row_occupied[row] &&
	! node->diag1_occupied[MaxN + col + row] &&
	! node->diag2_occupied[MaxN - col + row]) {
      *p = row;
      if (node->d == global->n)
	nq_Solution(global, node);
      else {
	node->row_occupied[row] = 1;
	node->diag1_occupied[MaxN + col + row]++;
	node->diag2_occupied[MaxN - col + row]++;
	whatToDo((z_NodeP) node, ref);
	node->row_occupied[row] = 0;
	node->diag1_occupied[MaxN + col + row]--;
	node->diag2_occupied[MaxN - col + row]--;
      }
    }
  node->d--;
  return 0;
} /* nq_BranchInPlace */

void initproc(void)
{
  /* This function is called once on every processing element. */

  z_bt1_Init();
  z_bt2_Init();
  
  nq_NodeTag = z_InstallClass(nq_Node, 0, 0, 0);
  nq_GlobalTag = z_InstallClass(nq_Global, 0, 0, 0);
#ifdef COUNT
  nq_TotalTag = z_InstallClass(nq_Total, 0, 0, 0);
#ifndef SEQUENTIAL
  totalJobNum = z_InstallJob(CollectTotalJob);
#endif /* SEQUENTIAL */
  /* On every processor, a local counter is allocated. */
  total = (nq_TotalP) z_NewNode(nq_TotalTag, sizeof(nq_Total));
  total->sol_count = 0;
#endif /* COUNT */
  nq_procset.bt_BranchInPlace = (z_bt_BranchProc)nq_BranchInPlace;
  nq_procsetnum = z_InstallProcSet((z_ProcSetP) &nq_procset);
} /* initproc */

static int Usage(const char *name)
{
  fprintf(stderr, "Usage: %s [-1chnprt] n\n\n", name);
  fprintf(stderr, "Options: -1  sequential\n");
  fprintf(stderr, "         -h  help (produce this message)\n");
  fprintf(stderr, "         -n  no output (implies time measurement)\n");
  fprintf(stderr, "         -p  PostScript output\n");
  /*fprintf(stderr, "         -r  restart\n");*/
  fprintf(stderr, "         -t  time measurement\n");
  fprintf(stderr, "n must be in the range 1..%d\n", MaxN);
  return 4;
} /* Usage */

int mmain(int argc, char **argv, char **envp)
{
  nq_GlobalP global;
  nq_NodeP root;
  int i;
  double time;
  enum outputType outputType = ot_normal;
  boolean showTime=0;
  boolean sequential=0;
  boolean restart=0;
  char c;

  while ((c=getopt(argc, argv, "1hnprt")) != -1)
    switch(c) {
    case 'c':
      outputType = characterGraphic;
      break;
    case 'n':
      outputType = noOutput;
      showTime = 1;
      break;
    case 'r':
      restart = 1;
      break;
    case 'p':
      outputType = postScript;
      sequential = 1;
      break;
    case 't':
      showTime = 1;
      break;
    case '1':
      sequential = 1;
      break;
    default:
      return Usage(argv[0]);
    }
  
  if (argc != optind+1) return Usage(argv[0]);
  
  global = (nq_GlobalP) z_NewNode(nq_GlobalTag, sizeof(nq_Global));
  if (!global) exit(1);
  global->n = atoi(argv[optind]);
  global->outputType = outputType;
  if (global->n < 1 || global->n > MaxN) return Usage(argv[0]);
  
  root = nq_NewNode();
  root->d = 0;
  for (i = 0; i < global->n; i++) root->row_occupied[i] = 0;
  for (i = 0; i < 2*MaxN; i++)
    root->diag1_occupied[i] = root->diag2_occupied[i] = 0;

  switch (outputType) {
  case characterGraphic:
  case postScript:
    z_enumps_Open("/tmp/nq");
    break;
  default:
    ;
  }
  
  time = z_Gtime();
  if (sequential)
    z_bt_EnumerateDepthFirst(nq_procsetnum, (z_bt_GlobalP) global,
			     (z_NodeP)root);
  else
    z_bt_EnumerateDFExp(nq_procsetnum, (z_bt_GlobalP) global,
			(z_NodeP)root, sequential);
  time = z_Gtime() - time;

#ifdef COUNT
#ifndef SEQUENTIAL
  /* Add the local solution counters of all processors to the
     counter on the front end. */
  z_StartParallelJob(totalJobNum, AddTotal);
  z_WaitWork();
#endif /* SEQUENTIAL */
  /* Now print the value of the counter on the front end. */
  fprintf(stderr, "Number of solutions: %ld\n", total->sol_count);
#endif /* COUNT */
  if (outputType == postScript) {
    z_enumps_ratio = 1;
    z_enumps_frame = 1;
    z_enumps_borderFraction = 1/18.0;
    z_PrintPS((z_NodeP)global, stdout,
	      (z_enumps_HeaderProc)nq_PSHeader,
	      (z_enumps_PictureProc)nq_PSPicture);
  }
  if (showTime)
    if (sequential)
      printf("Time:%7.2f (sequential) n = %d\n",
	     time, global->n);
    else
      printf("Time:%7.2f (%2d processor%s n = %d\n",
	     time, z_numprocs, z_numprocs>1?"s)":") ", global->n);
  return 0;
} /* mmain */
