/*--

   Copyright (c) 2004 Ambros Marzetta and David Bremner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

--*/
/* s_seq.c (sequence, implemented as doubly linked list or as array)

   Ambros Marzetta
   
 * $Source$
 * $Revision: 5167 $
 * $Date: 2004-10-20 13:20:33 -0300 (Wed, 20 Oct 2004) $
 * $Log$
 * Revision 1.7  2004/10/20 16:20:31  bremner
 * added copyright notice and GPL boilerplate
 *
 * Revision 1.6  2004/08/31 00:08:56  bremner
 * working on standalone install
 *
 * Revision 1.5  2004/08/25 21:01:45  bremner
 * libutil configuration
 * zram without mpi
 *
 * Revision 1.4  2002/04/04 15:55:07  bremner
 * *** empty log message ***
 *
 * Revision 1.3  2002/04/04 14:34:47  bremner
 * about to rip out a bunch of custom memory allocation code and replace
 * it with calls to memalloc. Because they are making me crazy
 *
 * Revision 1.2  2000/11/15 16:14:52  bremner
 *
 * committing patched version of zram source
 *
 * Revision 1.2  1997/04/25  08:41:11  marzetta
 * Added some type casts.
 *
 * Revision 1.1  1996/04/11  16:00:28  marzetta
 * Initial revision
 *
 */

#ifdef HAVE_CONFIG_H
#include "zram_config.h"
#endif

#define _S_SEQ_C_
#include "s_seq.h"
#include <stdio.h>
#include <stdlib.h>		/* malloc */
#include "z_par.h"		/* assert */

#ifdef USE_DMALLOC
#include <dmalloc.h>
#endif

#include "z_memalloc.h"

/* Types ------------------------------------------------------------- */

struct s_seqT {
  long numElem;			/* number of elements in the sequence */
  s_seq_typeT type;
};

/* Doubly linked list implementation --------------------------------- */

typedef struct dl_refT {
  struct dl_refT *next, *prev;
  s_seq_elemP elem;
} *dl_refP;

typedef struct {
  struct s_seqT seq;	     /* common part for all sequence implementations */
  struct dl_refT anchor;
} *dl_seqP;

#if 0
static dl_refP freeList;
static int freeListSize;
#endif

dl_seqP z_dl_New(void);
void z_dl_Free(dl_seqP l);
s_seq_elemP z_dl_GetObject(dl_seqP l, dl_refP where);
dl_refP z_dl_InsertFirst(dl_seqP l, s_seq_elemP x);
dl_refP z_dl_InsertLast(dl_seqP l, s_seq_elemP x);
dl_refP z_dl_InsertAfter(dl_seqP l, dl_refP r, s_seq_elemP x);
dl_refP z_dl_First(dl_seqP l);
dl_refP z_dl_Last(dl_seqP l);
void z_dl_ForAll(dl_seqP l, z_PtrProc whatToDo);
void z_dl_Delete(dl_seqP l, dl_refP ref);

dl_seqP z_dl_New(void)
{
  dl_seqP l;
  
  l = (dl_seqP) memalloc(sizeof(*l));
  assert(l);
  l->anchor.next = & l->anchor;
  l->anchor.prev = & l->anchor;
  l->anchor.elem = NULL;
  return l;
} /* z_dl_New */

void z_dl_Free(dl_seqP l)
{
  free(l);
} /* z_dl_Free */

s_seq_elemP z_dl_GetObject(dl_seqP l, dl_refP where)
{
  return where->elem;
} /* z_dl_GetObject */

dl_refP z_dl_InsertFirst(dl_seqP l, s_seq_elemP x)
{
  dl_refP ref;

  ref = (dl_refP) memalloc(sizeof(*ref));
  assert(ref);
  ref->elem = x;
  ref->prev = & l->anchor;
  ref->next = l->anchor.next;
  l->anchor.next->prev = ref;
  l->anchor.next = ref;
  return ref;
} /* z_dl_InsertFirst */

dl_refP z_dl_InsertLast(dl_seqP l, s_seq_elemP x)
{
  dl_refP ref;

  ref = (dl_refP) memalloc(sizeof(*ref));
  assert(ref);
  ref->elem = x;
  ref->next = & l->anchor;
  ref->prev = l->anchor.prev;
  l->anchor.prev->next = ref;
  l->anchor.prev = ref;
  return ref;
} /* z_dl_InsertLast */

dl_refP z_dl_InsertAfter(dl_seqP l, dl_refP r, s_seq_elemP x)
{
/* Insert x into the sequence l after r. */
  assert(0);
  return NULL;
} /* z_dl_InsertAfter */

dl_refP z_dl_First(dl_seqP l)
{
  return l->anchor.next;
} /* z_dl_First */

dl_refP z_dl_Last(dl_seqP l)
{
  return l->anchor.prev;
} /* z_dl_Last */

void z_dl_ForAll(dl_seqP l, z_PtrProc whatToDo)
{
  dl_refP ref;

  for (ref = l->anchor.next; ref != &l->anchor; ref = ref->next)
    whatToDo(ref->elem);
} /* z_dl_ForAll */

void z_dl_Delete(dl_seqP l, dl_refP ref)
{
  ref->prev->next = ref->next;
  ref->next->prev = ref->prev;
  memfree(ref);
} /* z_dl_Delete */

/* Implementation table ---------------------------------------------- */

typedef s_seqP (*NewProc)(void);
static NewProc New[] = {(NewProc)z_dl_New};
     
typedef void (*FreeProc)(s_seqP s);
static FreeProc Free[] = {(FreeProc)z_dl_Free};

typedef s_seq_elemP (*GetProc)(s_seqP l, s_seq_refP where);
static GetProc Get[] = {(GetProc)z_dl_GetObject};

typedef s_seq_refP (*InsertProc)(s_seqP l, s_seq_elemP x);
static InsertProc InsertFirst[] = {(InsertProc)z_dl_InsertFirst},
                  InsertLast[] = {(InsertProc)z_dl_InsertLast};

typedef s_seq_refP (*InsertAfterProc)(s_seqP l, s_seq_refP r, s_seq_elemP x);
static InsertAfterProc InsertAfter[] = {(InsertAfterProc)z_dl_InsertAfter};

typedef s_seq_refP (*RefProc)(s_seqP l);
static RefProc First[] = {(RefProc)z_dl_First},
               Last[] = {(RefProc)z_dl_Last};

typedef void (*DeleteProc)(s_seqP l, s_seq_refP ref);
static DeleteProc Delete[] = {(DeleteProc)z_dl_Delete};

typedef void (*ForAllProc)(s_seqP l, z_PtrProc whatToDo);
static ForAllProc ForAll[] = {(ForAllProc)z_dl_ForAll};
     
/* Procedures (general)----------------------------------------------- */

s_seqP s_seq_New(s_seq_typeT type)
{
/* Allocate an empty list. Choose an implementation
   of type type. */
  s_seqP s;
  
  assert(type < s_seq_numtypes);
  s = New[type]();
  s->numElem = 0;
  s->type = type;
  return s;
} /* s_seq_New */


void s_seq_Free(s_seqP l)
{
/* Deallocate the sequence l, which must be empty. */
  assert(l->numElem == 0);
  Free[l->type](l);
} /* s_seq_Free */


s_seq_elemP s_seq_GetObject(s_seqP l, s_seq_refP where)
{
/* Retrieve the object referenced by where. */
  assert(l->type < s_seq_numtypes);
  return Get[l->type](l, where);
} /* s_seq_GetObject */


s_seq_refP s_seq_InsertFirst(s_seqP l, s_seq_elemP x)
{
/* Insert x into the sequence l (at the beginning). */
  l->numElem ++;
  return InsertFirst[l->type](l, x);
} /* s_seq_InsertFirst */


s_seq_refP s_seq_InsertLast(s_seqP l, s_seq_elemP x)
{
/* Insert x into the sequence l (at the end). */
  l->numElem ++;
  return InsertLast[l->type](l, x);
} /* s_seq_InsertLast */


s_seq_refP s_seq_InsertAfter(s_seqP l, s_seq_refP r, s_seq_elemP x)
{
/* Insert x into the sequence l after r. */
  l->numElem ++;
  return InsertAfter[l->type](l, r, x);
} /* s_seq_InsertAfter */


long s_seq_numElements(s_seqP l)
{
/* Return the number of elements in l. */
  return l->numElem;
} /* s_seq_numElements */


s_seq_refP s_seq_First(s_seqP l)
{
/* Return a reference to the first element of the sequence.
   The sequence must not be empty. */
  assert(l->numElem > 0);
  return First[l->type](l);
} /* s_seq_First */


s_seq_refP s_seq_Last(s_seqP l)
{
/* Return a reference to the last element of the sequence.
   The sequence must not be empty. */
  assert(l->numElem > 0);
  return Last[l->type](l);
} /* s_seq_Last */

void s_seq_ForAll(s_seqP l, z_PtrProc whatToDo)
{
  /* Call whatToDo for all elements of the sequence. WhatToDo must not
     access the sequence. */
  ForAll[l->type](l, whatToDo);
} /* s_seq_ForAll */

void s_seq_Delete(s_seqP l, s_seq_refP ref)
{
/* Delete the element referenced by ref. */
  l->numElem --;
  Delete[l->type](l, ref);
} /* s_seq_Delete */
